#ifdef NQ_HEADER_BEGIN_do_not_delete
#endif  // NQ_HEADER_END_do_not_delete
/*
 * i2c sniffer with packet buffering based on 
 * i2c sniffer by storepeter (C) 2022 Beerware
 * sends sampled data on rs232 @ 2Mbit for further analysis on linux
 *
 * SCL PB1 - pin 9 on Arduino   to A5 SCl on arduino being monitored
 * SDA PB0 - pin 8 on Arduino   to A4 SDA on arduino being monitored
 *
 * updated by Alter (c) 2023:
 *   buffered sending (not immediately in real-time)
 *   optional timestamping
 *   optional ASCII dump
 *
 * dump to memory with defered send to client 
 *
 */
#define  RECORD_SAMPLING_TIME 1
//#define DUMP_TIMESTAMP
#define DUMP_ASCII
#ifdef ARDUINO
// I am using my own build system,
// here is what is needed to make this compile from the Arduino IDE
# define NQ_PUBLIC
# if   defined(__AVR_ATmega8__)
#  define _BV_URSEL _BV(URSEL)  // atmega8 need to set URSEL to modify UCSRC
# else                          // Newer atmega168/atmega328p
#  define _BV_URSEL 0           // atmega8 need to set URSEL to modify UCSRC
#  define TXEN   TXEN0
#  define U2X    U2X0
#  define UBRRL  UBRR0L
#  define UCSRA  UCSR0A
#  define UCSRB  UCSR0B
#  define UCSRC  UCSR0C
#  define UCSZ0  UCSZ00
#  define UCSZ1  UCSZ01
#  define UDRE  UDRE0
#  define UDR    UDR0
# endif
#else
# include "nq.h"
#endif
 
#include "i2c-sniff.h"
#include "usb-i2c.h"

/*
uint8_t buff[256]
    __attribute__((aligned(256)))       // aligned to 256 so r4,r5=my_log -> r5=index
    ;
    */
extern uint8_t iobuffer;
extern uint16_t i2c_addr;

uint8_t *buff = &iobuffer[0];
uint8_t *serial_buff = &iobuffer[256];

#define SDApin  8
#define SCLpin  9
 
#define SCL _BV(1)
#define SDA _BV(0)

/* 
#ifdef DUMP_TIMESTAMP
const uint8_t my_log[256]
    __attribute__((aligned(256)))       // aligned to 256 so r2,r3=my_log -> r3=index
    __attribute__((section(".fini9"))) = {
        0 + STOP, 1 + STOP, 2 + STOP, 2 + STOP, 3 + STOP, 3 + STOP, 3 + STOP, 3 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, 4 + STOP, // 0x00
        5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, 5 + STOP, // 0x10
        6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, // 0x20
        6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, 6 + STOP, // 0x30
        7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, // 0x40
        7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, // 0x50
        7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, // 0x60
        7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, 7 + STOP, // 0x70
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0x80
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0x90
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0xa0
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0xb0
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0xc0
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0xd0
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, // 0xe0
        8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP, 8 + STOP  // 0xf0
};

static inline uint8_t do_my_log(int8_t offset)
{
        uint8_t value;
        asm volatile ("mov r31,r3\n\t" "mov r30,%1\n\t" "lpm %0,z\n\t":"=r" (value)
                      :"r"(offset)
            );
        return value;
}
#endif //DUMP_TIMESTAMP
*/
 
NQ_PUBLIC void _nqprintc(char ch)
{
    while (!(UCSRA & _BV(UDRE))) ;
    UDR = ch;
}
 
#ifdef SAMPLE_CHK
#define timer0_reset()  do { TCNT0 = SAMPLE_CHK; } while (0)
#else
#define timer0_reset()  while (0)
#endif
int run_sniff()
{
    register uint8_t i2c asm("r16");
    register uint8_t prev asm("r17");
    uint8_t bit;
    uint8_t data;
    //uint8_t c_time;
    uint8_t c_state;
    bool    bPause;
/*
#ifdef DUMP_TIMESTAMP
    register const uint8_t *my_log_pt asm("r2");    // r2=lsb r3=msb
#endif //DUMP_TIMESTAMP
*/
    //register uint8_t *byte_log_pt asm("r4");    // r4=lsb r5=msb
    //register uint8_t idx asm("r5");
    uint8_t a = 0;
    uint8_t idx = 0;
    uint8_t jdx = 0;
    uint8_t i;
    uint16_t skipped = 0;
    uint16_t last_len = 0;
/*
#ifdef DUMP_TIMESTAMP
    //my_log_pt = &(my_log[0]);
#endif //DUMP_TIMESTAMP
*/
    //byte_log_pt = &(buff[0]);
#ifdef SAMPLE_CHK
    register uint8_t max;
    max = 0;
    TCCR0B = _BV(CS00);     // Timer0 8bit  prescaler=18  16 Mhz clk
#endif
    TCCR1B = _BV(CS12);     // Timer1 16bit prescaler=256 16 usec clk

    //Serial.begin(115200);  // max for ch341-uart, assume already started
    //UCSRA |= _BV(U2X);      // double speed
    //UCSRB |= _BV(TXEN);
    //UCSRC = _BV_URSEL | _BV(UCSZ1) | _BV(UCSZ0);    // config USART; 8N1 (BV_URSEL for atmega8
    //UBRRL = 0;              // 2 Mbit/sec

    const char *cpt = "i2c_sniff\n\r";
    while (*cpt)
        _nqprintc(*cpt++);

    bit = 0;
    data = 0;
    data_state = 0;
    //c_time = 0;
    c_state = 0;
    i2c = PINB & 0x03;  // SCL=PB0, SDA=PB1
    prev = i2c;             // SCL=PB1, SDA=PB0
    last_len = 0;
    last_pos = 0;
    timer0_reset();
    while (1) {
#ifdef SAMPLE_CHK
        if (TCNT0 > max) {   // will settle after a few cycles
            max = TCNT0;
/*
#ifdef DUMP_TIMESTAMP
            UDR = max;      // get it out there hints where it was created too
#endif //DUMP_TIMESTAMP
*/
        }
        timer0_reset();
#endif
        prev = i2c;
        i2c = PINB;     // SCL=PB1, SDA=PB0
        
        //if(i2c == prev)
        //  continue;

        if (prev & SCL) {   // SCL = """\__
            if ((i2c & SCL) == 0) {     // SCL pulled down
                if (bit == 9) {
                    if (prev & SDA) {   // ACK
                        //data |= NACK; // end transmission
                        data_state |= NACK;

                        // don't include addr in length
                        i2c_addr = ((uint16_t)buff[last_pos] & 0xfe) >> 1;
                        if((i2c_addr & 0xf8) == 0xf0) {
                            last_len--;
                        }
                        last_len--;

                        buff[idx++] = UI2C_RAW_ERR_PREFIX;
                        buff[idx++] = last_len ? UI2C_2W_ERR_DATA_NACK : UI2C_2W_ERR_ADDR_NACK;

                        buff[last_pos] = last_len;
                        last_len = 0;
                        last_pos = idx; // reserved place for data length
                        idx++;

                    } else {
                        //data |= ACK; // continue transmission
                        data_state |= ACK;
                    }
                    //UDR = data;
                    /*
                    *byte_log_pt = data;
                    idx++;
                    */
                    //buff[idx++] = data;
                    skipped=0;

                    bit = 1;
                    continue;
                }
                data <<= 1;
                //data &= 0x0f;       // 4 lsb contains half byte // use 8 bits
                data |= i2c & SDA;
                //if (bit == 4) 
                if (bit == 8) 
                {
                    //data |= c_state;        // first START then DATA
                    data_state |= c_state;
                    //UDR = data;
                    /*
                    *byte_log_pt = data;
                    idx++;
                    */
                    buff[idx++] = data;
                    skipped=0;

                    last_len++;

                    c_state = DATA;
                }
                bit++;
                continue;
            }       // SCL unchanged high, SDA must have changed
            if (prev & SDA) {   // SDA was 1
                if ((i2c & SDA) == 0) {     // SDA is now 0
                    // SDA was pulled low START condition
/*
#ifdef DUMP_TIMESTAMP
                    if (c_time) {   // send idle time if there is one
                        UDR = c_time;
                    }
#endif //DUMP_TIMESTAMP
*/
                    /*
                    *byte_log_pt = START;
                    idx++;
                    */
                    //buff[idx++] = START;
                    skipped=0;

                    //c_time = 0;     // clear timer1 (16 bit)
                    c_state = START;
                    bit = 1;        // falling scl advances to 1,while SDA=0

                    if(last_len)
                    {
                        last_pos = idx; // reserved place for data length
                        idx++;
                        last_len=0;
                    }

                    continue;;
                }
            } else {        // SDA was 0
                if (i2c & SDA) {    // SDA is now 1
                    // SDA went high, STOP condition
/*
#ifdef DUMP_TIMESTAMP
                    UDR = do_my_log(TCNT1L);
#endif //DUMP_TIMESTAMP
*/
                    /*
                    *byte_log_pt = STOP;
                    idx++;
                    */
                    //buff[idx++] = STOP | bit;
                    //buff[idx++] = STOP | bit;
                    data_state = STOP;

                    // don't include addr in length
                    i2c_addr = ((uint16_t)buff[last_pos] & 0xfe) >> 1;
                    if((i2c_addr & 0xf8) == 0xf0) {
                        last_len--;
                    }
                    last_len--;

                    buff[last_pos] = last_len;
                    last_len = 0;
                    last_pos = idx; // reserved place for data length
                    idx++;

                    //c_time = 0;     // clear timer1 (16 bit)
                    //timer0_reset(); // sample time only recorded when active
                    continue;
                }
            }
            // SDA unchanged
        }
        //if (c_time == 0) {
        //     TCNT1 = 0;      // clear timer1 (16 bit)
        //}
        // we are not busy so just figure out what c_time is now
        union {
            uint8_t u8[2];
            uint16_t u16;
        } count;
        count.u16 = TCNT1;

        if (count.u8[1] > 0xf8 || (uint8_t)idx == (uint8_t)(jdx-127) || skipped>16384) {
            //Serial.println(IDLE, HEX);
#ifdef DUMP_ASCII
            bPause = (jdx != idx);
            if(bPause)
            {
                Serial.println("--");
            }
            while(jdx != idx)
            {
                Serial.println(buff[++jdx], HEX);
            }
            if(bPause)
            {
                Serial.flush();
            }
#else //DUMP_ASCII
            while(jdx != idx)
            {
                _nqprintc(buff[++jdx]);
            }
#endif //DUMP_ASCII
            skipped=0;
            //Serial.println("--");

            /**** check exit request ****/
            i = 0;
            while (Serial.available() > 0)
            {
                serial_buff[i] = Serial.read();
                if(i<16)
                {
                    i++;
                }
            }
            if(i=5 && !memcmp(serial_buff, CMD_RAW_TO_MANUAL, 5)) {    // mode=xxxxx
                break;
            }

        } else {
            skipped++;
        }

        if (count.u8[1] > 0xf8) {
/*
#ifdef DUMP_TIMESTAMP
            UDR = IDLE + 8 + 8;
#endif //DUMP_TIMESTAMP
*/
            //c_time = 0;     // clear timer1 (16 bit)
            timer0_reset(); // sample time only recorded when active
            continue;
        }
/*
#ifdef DUMP_TIMESTAMP
        if (count.u8[1]) {
            c_time = IDLE - STOP + 8 + do_my_log(count.u8[1]);
        } else {
            c_time = IDLE - STOP + do_my_log(count.u8[0]);
        }
#endif //DUMP_TIMESTAMP
*/
        timer0_reset(); // sample time only recorded when active
    }
}
